// jit.h

// JIT opcodes
//
// Opcode suffixes
// r_r      REAL vec and REAL vec
// r_r1     REAL vec and REAL scalar
// r1_r     REAL scalar and REAL vec
// r1_r1    REAL scalar and REAL scalar
//
// The order within blocks of the opcodes below matters, see
// genjitIntBin() and adjustForTypes() and friends.
//
// If you add an opcode, extend JIT_OPCODE_NAMES.

typedef enum JIT_OPCODE {
    JIT_endop,          // marks end of opcode sequence

    JIT_push_r,         // push REAL vec
    JIT_push_i,         // push INTEGER vec
    JIT_Cpush_r,        // as above with with stack overflow check
    JIT_Cpush_i,

    JIT_pushsym_r,      // push REAL SYMSXP
    JIT_pushsym_i,      // push INTEGER SYMSXP
    JIT_Cpushsym_r,     // as above with with stack overflow check
    JIT_Cpushsym_i,

    JIT_math1_r,        // math function of one real vector argument
    JIT_math1_r1,       // ditto, but for scalar argument
    JIT_math1_i,        // math function of one integer vector argument
    JIT_math1_i1,       // ditto, but for scalar argument

    // e.g. JIT_subset_r_i1 is x[i] where x is a real vec and i is an int scalar

    JIT_subset_r_r, JIT_subset_r_r1,
    JIT_subset_r_i, JIT_subset_r_i1,
    JIT_subset_i_r, JIT_subset_i_r1,
    JIT_subset_i_i, JIT_subset_i_i1,

    JIT_Csubset_r_r, JIT_Csubset_r_r1,   // as above but with index range check
    JIT_Csubset_r_i, JIT_Csubset_r_i1,
    JIT_Csubset_i_r, JIT_Csubset_i_r1,
    JIT_Csubset_i_i, JIT_Csubset_i_i1,

    // currently we only use assign ops where both sides have the same length

    JIT_assign_r_r, JIT_assign_r_r1, JIT_assign_r1_r, JIT_assign_r1_r1,
    JIT_assign_r_i, JIT_assign_r_i1, JIT_assign_r1_i, JIT_assign_r1_i1,
    JIT_assign_i_r, JIT_assign_i_r1, JIT_assign_i1_r, JIT_assign_i1_r1,
    JIT_assign_i_i, JIT_assign_i_i1, JIT_assign_i1_i, JIT_assign_i1_i1,

    JIT_add_r_r, JIT_add_r_r1, JIT_add_r1_r, JIT_add_r1_r1,
    JIT_add_r_i, JIT_add_r_i1, JIT_add_r1_i, JIT_add_r1_i1,
    JIT_add_i_r, JIT_add_i_r1, JIT_add_i1_r, JIT_add_i1_r1,
    JIT_add_i_i, JIT_add_i_i1, JIT_add_i1_i, JIT_add_i1_i1,

    JIT_sub_r_r, JIT_sub_r_r1, JIT_sub_r1_r, JIT_sub_r1_r1,
    JIT_sub_r_i, JIT_sub_r_i1, JIT_sub_r1_i, JIT_sub_r1_i1,
    JIT_sub_i_r, JIT_sub_i_r1, JIT_sub_i1_r, JIT_sub_i1_r1,
    JIT_sub_i_i, JIT_sub_i_i1, JIT_sub_i1_i, JIT_sub_i1_i1,

    JIT_mul_r_r, JIT_mul_r_r1, JIT_mul_r1_r, JIT_mul_r1_r1,
    JIT_mul_r_i, JIT_mul_r_i1, JIT_mul_r1_i, JIT_mul_r1_i1,
    JIT_mul_i_r, JIT_mul_i_r1, JIT_mul_i1_r, JIT_mul_i1_r1,
    JIT_mul_i_i, JIT_mul_i_i1, JIT_mul_i1_i, JIT_mul_i1_i1,

    // div always return REAL even when args are INTs,
    // for compatibility with existing R code in arithmetic.c.
    JIT_div_r_r, JIT_div_r_r1, JIT_div_r1_r, JIT_div_r1_r1,
    JIT_div_r_i, JIT_div_r_i1, JIT_div_r1_i, JIT_div_r1_i1,
    JIT_div_i_r, JIT_div_i_r1, JIT_div_i1_r, JIT_div_i1_r1,
    JIT_div_i_i, JIT_div_i_i1, JIT_div_i1_i, JIT_div_i1_i1,

    // pow always return REAL even when args are INTs
    JIT_pow_r_r, JIT_pow_r_r1, JIT_pow_r1_r, JIT_pow_r1_r1,
    JIT_pow_r_i, JIT_pow_r_i1, JIT_pow_r1_i, JIT_pow_r1_i1,
    JIT_pow_i_r, JIT_pow_i_r1, JIT_pow_i1_r, JIT_pow_i1_r1,
    JIT_pow_i_i, JIT_pow_i_i1, JIT_pow_i1_i, JIT_pow_i1_i1,

    JIT_mod_r_r, JIT_mod_r_r1, JIT_mod_r1_r, JIT_mod_r1_r1,
    JIT_mod_r_i, JIT_mod_r_i1, JIT_mod_r1_i, JIT_mod_r1_i1,
    JIT_mod_i_r, JIT_mod_i_r1, JIT_mod_i1_r, JIT_mod_i1_r1,
    JIT_mod_i_i, JIT_mod_i_i1, JIT_mod_i1_i, JIT_mod_i1_i1,

    JIT_idiv_r_r, JIT_idiv_r_r1, JIT_idiv_r1_r, JIT_idiv_r1_r1,
    JIT_idiv_r_i, JIT_idiv_r_i1, JIT_idiv_r1_i, JIT_idiv_r1_i1,
    JIT_idiv_i_r, JIT_idiv_i_r1, JIT_idiv_i1_r, JIT_idiv_i1_r1,
    JIT_idiv_i_i, JIT_idiv_i_i1, JIT_idiv_i1_i, JIT_idiv_i1_i1,

    JIT_last,           // for sanity checking, never executed by JIT machine
} JIT_OPCODE;

extern unsigned globalJitDirective;
extern int jitGenDepth;
extern int jitDepth;
extern int jitUnresolved;
extern Rboolean jitCompileAborted;
SEXP do_jit(SEXP call, SEXP op, SEXP args, SEXP env);
SEXP evalJIT(const SEXP e);
SEXP getExpBeforeItWasJitted(const SEXP s);
void decJitUnresolved(void);
void ForwardJitNodes(SEXP *pforwarded_nodes);
void genjitAssign(const SEXP sym, SEXP location, const SEXP value, const SEXP env, const char msg[]);
void genjitIntBinAux(JIT_OPCODE opcode, const SEXP x, const SEXP y, const SEXP resultTemplate);
void genjitMath1(double(* const func)(double), const SEXP argTemplate, const SEXP ansTemplate);
void genjitPush(JIT_OPCODE opcode, SEXP e);
void genjitPushsym(SEXP location, SEXP env);
void genjitRealBin(ARITHOP_TYPE oper, const SEXP x, const SEXP y);
void genjitSubset(const SEXP x, const SEXP subset);
void jitEpilog(SEXP e, const SEXP resultTemplate, const SEXP op);
void jitInit(void);
void jitOff(Rboolean issueWarning);
void jitProlog(const SEXP e, const SEXP op, const char msg[]);
void pushJitState(SEXP call);
void popJitState(SEXP call);
void printJit(const SEXP s);
void checkJitBindingVal(const SEXP b, const SEXP val);
void checkJitSymBindingVal(const SEXP sym, const SEXP val);
void checkJitRemove(const SEXP name);
void disallowIfJitting(const char s[]);
#if 1 // RA_TODO these will disappear
void traceBinding(const SEXP b, const SEXP val);
void traceSymBinding(const SEXP sym, const SEXP val);
#endif
void checkLocationConsistency(SEXP e, SEXP env, SEXP loc, SEXP val);
void checkLocalBindingForJit(const SEXP sym, const SEXP val, const SEXP env);
Rboolean isAssignmentJittable(SEXP rhs);
void jitEnterLoop(const SEXP s);
void jitExitLoop(const SEXP s);
CCODE getCfun(const SEXP s);
void cannotJitCompile(const SEXP e);

// these inline functions are for efficiency without clutter

static R_INLINE Rboolean jitCompiling(void)
{
    return jitGenDepth && !jitCompileAborted;
}

static R_INLINE void genjitIntBin(JIT_OPCODE opcode, const SEXP x,
                                   const SEXP y, const SEXP resultTemplate)
{
    if (jitGenDepth)
        genjitIntBinAux(opcode, x, y, resultTemplate);
}
