% File src/library/utils/man/jit.Rd

\name{jit}
\alias{jit}
\title{Just-in-time compilation}
\description{
  Enable just-in-time (JIT) compilation of a block of R code.
  Arithmetic in JIT code is usually faster.
}
\usage{
  jit(jit = NA, trace = 0)
}
\arguments{
  \item{jit}{
    \strong{NA} make no changes but return the current JIT state (default)

    \strong{1}  start JIT block

    \strong{0}  end JIT block.
        Using \code{jit(0)} is often unnecessary because an implicit \code{jit(0)} is
        performed when the function returns.
  }
  \item{trace}{
    \strong{0} silent (default)

    \strong{1} show which expressions are compiled to JIT code as they are compiled.

    \strong{2} show compilation details

    \strong{3} show code generation details.
  }
}
\details{
\strong{Nomeclature}

A \emph{JIT block} is the R code between the \code{jit(1)} and 
\code{jit(0)} or end of the function.

\emph{Jitted code} is R code that has been compiled by the
just-in-time compiler (to a form where it will execute more
quickly).
Not all code in a JIT block is necessarily jitted.

A \emph{jitted variable} is a variable used in jitted code.

\strong{What gets jitted?}

Currently, the following operations are jitted, and only if they
are in a \code{for}/\code{while}/\code{repeat} loop in a JIT block:

\itemize{
\item Basic binary arithmetic operations on reals, integers, 
  and logicals:\cr
  \code{+ - & / ^}\cr
  Both arguments of the binary operator must have the same
  length, or one of them must have length 1 (general R recycling
  is not jitted).

\item Scalar subscripts of double and integer vectors: \code{x[i]}

\item  Simple assignment: \code{x <- rhs}\cr
  But not subscripted assignment: \code{x[i] <- rhs}

\item Intrinsic math functions of one argument: \code{sin(x)} and friends
}
Rather than remembering these rules, it may be simpler to
use \code{trace=1} to see what was jitted.
Code in a JIT block that is not actually jitted will run as
normal R code and coexist happily with jitted code.
Functions called from the JIT block are not jitted.

The above list will be expanded once the basic structure
of this early version of the jitter is solid.

\strong{How does jitted code differ semantically from non-jitted code?}

[This section will be edited to make it more clear].

R will not allow you to change the length or type of a jitted variable.
The idea is to keep the evironment stable so the code can be efficient.

In a JIT block, avoid \code{NA}s:
\itemize{
  \item Logical and integer \code{NA}s cause unpredictable results.\cr
        More precisely, if the \code{NA} is in a jitted expression it 
        will be treated as \code{INT_MIN}, which is R's internal 
        representation of logical and integer \code{NA}s.
        If the expression is not jitted even though it is in a jit block, 
        the \code{NA} will be handled with R's usual semantics.
  \item Use \code{NaN} instead of \code{NA} in jitted real expressions.\cr
        Although real \code{NA}s are handled with R's usual semantics
        in jitted code, remember that a plain "\code{NA}" is treated
        as \emph{logical} by R i.e. as \code{INT_MIN} in jitted code.\cr
        In the example standard R code below, the logical NA gets 
        converted to real because
        1.2 is real:\cr
\code{1.2 + NA -> NA}\cr
        And in jitted code:\cr
\code{1.2 + NA -> 1.2 + INT_MIN -> -2.15e+09}\cr
}
In a JIT block, avoid out-of-range subscripts.
Out-of-range subscripts in jitted code cause unpredictable results.
More precisely, if the out-of-range index is in a jitted expression
it will be caught with a runtime error message.
If the out-of-range index is in a JIT block but in an expression that
is not actually jitted,
the out-of-range index will be handled with R's usual semantics.
But note that an out-of-range subscript will never corrupt memory ---
jitted code is still R code, not C code!

\strong{Limits of the current implementation}

The currently known limitations are listed below.
These need to be taken care of before the
jitter becomes truly useful.

\itemize{
    \item You can only jit one function at any time.
       A call to \code{jit} in a function that is called by a 
       jitted function is ignored, with a warning message.  

    \item Not much gets jitted. See \strong{What gets jitted?} above.

    \item The speed improvement is not yet impressive...
}
}
\value{
 A four element integer vector.
 The first two elements give the current state of the above two arguments.
 The remaining two elements are for internal use.
}
\examples{
foo <- function(N) {
    x <- 0
    for (i in 1:N)
       x <- x + 1
    x
}
foojit <- function(N) {
    jit(1)          # turn on just-in-time compilation
    x <- 0
    for (i in 1:N)
       x <- x + 1   # an arithmetic expression in a loop, so will be jitted
    x
}
cat("Time ratio",
    system.time(foo(1e5))[1] / system.time(foojit(1e5))[1],
    "\n")
}
\keyword{utilities}
