// jit.c: just-in-time compiler for R

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif
#include "Defn.h"
#include "Print.h"
#include "jit.h"
#include "printsxp.h"

#define DEBUG_JIT       1   // 0 no runtime checks (don't use yet)
                            // 1 Dassert (use in production build for now)
                            // 2 print jit code execution
                            // 3 gory details

#define DEBUG_SYM       1   // do expensive symbol binding consistency checks

#define MAX_JITS        101 // 1+max number of compiled jit expressions

#define MAX_JIT_EXP_LEN 401 // 1+max jit expression len (+1 is for JIT_endop)
                            // e.g. for "1+2" this needs 4 (push,push,add,endop)

// We use "Dassert" where speed is critical and "assert" where speed
// is not critical or the assertion is absolutely necessary.
#if DEBUG_JIT
    #define Dassert(x) assert(x)
#else
    #define Dassert(x)
#endif

#define STRLEN          80
#define CSEXP           const SEXP
#define NBR_ELEMS(x)    (sizeof(x) / sizeof((x)[0]))
#define printf          Rprintf

typedef double(* FUNC_TYPE)();

typedef struct JIT_OP {  // jit instruction (RA_TODO could be unionized)
    JIT_OPCODE  opcode;  // e.g. JIT_push_r
    SEXP        operand; // e.g. SEXP being pushed by JIT_push_r
                         //      or pointer to symbol binding for JIT_pushsym_r
    FUNC_TYPE   func;    // math function e.g. sin()
    SEXP        result;  // buffer to store the result e.g. for JIT_add_r_r
#if DEBUG_SYM
    SEXP        sym;     // symbol for JIT_pushsym_r \__for consistency check
    SEXP        env;     // env    for JIT_pushsym_r /
#endif
} JIT_OP;

typedef struct JIT_RECORD {
    SEXP    original;       // pointer to SEXP holding exp before it was jitted
    SEXP    ans;            // buf to store the result of evaluating jit exp
    JIT_OP  ops[MAX_JIT_EXP_LEN];
} JIT_RECORD;

// Values for bits in globalJitDirective.
// All bits will be 0 unless JIT_ON is 1.
//
// About JIT_RANGE_CHECK: by default this bit is 1.  When it is 0, no
// range checks are performed in evalJIT.   But this currently gives almost
// no speed advantage, so for now we always keep it set to 1.  If we
// develop the code to the point where evalJIT is significantly faster,
// we can extend do_jit to allow the user to clear the bit, at his own risk.

static const unsigned JIT_ON          = 0x1;
static const unsigned JIT_RANGE_CHECK = 0x2;

unsigned globalJitDirective;    // set by user with R jit() command
static int jitTrace;            // copy of R jit() trace parameter

// following scalars track progress through the R exp being evaluated by eval()

static unsigned jitDirective;  // jit directive at current func nesting level

int jitDepth;   // 0  do not generate jit code
                // 1  jitDirective != 0 but not in an R loop
                // >1 jitDirective != 0 and in an R loop
                // this variable will always be 0 if jitDirective==0

int jitGenDepth;  // depth in jittable eval tree
                  // if >0 then must gen jit code for the current eval()
                  // this variable will always be 0 if jitDepth <= 1

int jitUnresolved;      // must be zero for genex to be accepted as valid

Rboolean jitCompileAborted;

static RPRSTACK *pendingPromisesAtJitProlog;

static SEXP genex;      // the jit expression currently being generated
static int  ngenex;     // ops in genex from 0 to ngenex-1 are used

static SEXP jitted[MAX_JITS]; // jitted expressions i.e. array of JITSXPs
static int  njitted;          // SEXPs in jitted from 0 to njitted-1 are used

#define MAX_STACK_LEN MAX_JIT_EXP_LEN
static SEXP stack[MAX_STACK_LEN];       // jit exec stack, stack[0] isn't used
static int  istack = 0;                 // index into exec stack

static const char *jitFuncName;         // name of func that called jit(1)

static const char *JIT_OPCODE_NAMES[] = {
    "JIT_endop",

    "JIT_push_r",  "JIT_push_i",
    "JIT_Cpush_r", "JIT_Cpush_i",

    "JIT_pushsym_r",  "JIT_pushsym_i",
    "JIT_Cpushsym_r", "JIT_Cpushsym_i",

    "JIT_math1_r",
    "JIT_math1_r1",
    "JIT_math1_i",
    "JIT_math1_i1",

    "JIT_subset_r_r", "JIT_subset_r_r1",
    "JIT_subset_r_i", "JIT_subset_r_i1",
    "JIT_subset_i_r", "JIT_subset_i_r1",
    "JIT_subset_i_i", "JIT_subset_i_i1",

    "JIT_Csubset_r_r", "JIT_Csubset_r_r1",
    "JIT_Csubset_r_i", "JIT_Csubset_r_i1",
    "JIT_Csubset_i_r", "JIT_Csubset_i_r1",
    "JIT_Csubset_i_i", "JIT_Csubset_i_i1",

    "JIT_assign_r_r", "JIT_assign_r_r1", "JIT_assign_r1_r", "JIT_assign_r1_r1",
    "JIT_assign_r_i", "JIT_assign_r_i1", "JIT_assign_r1_i", "JIT_assign_r1_i1",
    "JIT_assign_i_r", "JIT_assign_i_r1", "JIT_assign_i1_r", "JIT_assign_i1_r1",
    "JIT_assign_i_i", "JIT_assign_i_i1", "JIT_assign_i1_i", "JIT_assign_i1_i1",

    "JIT_add_r_r", "JIT_add_r_r1", "JIT_add_r1_r", "JIT_add_r1_r1",
    "JIT_add_r_i", "JIT_add_r_i1", "JIT_add_r1_i", "JIT_add_r1_i1",
    "JIT_add_i_r", "JIT_add_i_r1", "JIT_add_i1_r", "JIT_add_i1_r1",
    "JIT_add_i_i", "JIT_add_i_i1", "JIT_add_ii_1", "JIT_add_i1_i1",

    "JIT_sub_r_r", "JIT_sub_r_r1", "JIT_sub_r1_r", "JIT_sub_r1_r1",
    "JIT_sub_r_i", "JIT_sub_r_i1", "JIT_sub_r1_i", "JIT_sub_r1_i1",
    "JIT_sub_i_r", "JIT_sub_i_r1", "JIT_sub_i1_r", "JIT_sub_i1_r1",
    "JIT_sub_i_i", "JIT_sub_i_i1", "JIT_sub_ii_1", "JIT_sub_i1_i1",

    "JIT_mul_r_r", "JIT_mul_r_r1", "JIT_mul_r1_r", "JIT_mul_r1_r1",
    "JIT_mul_r_i", "JIT_mul_r_i1", "JIT_mul_r1_i", "JIT_mul_r1_i1",
    "JIT_mul_i_r", "JIT_mul_i_r1", "JIT_mul_i1_r", "JIT_mul_i1_r1",
    "JIT_mul_i_i", "JIT_mul_i_i1", "JIT_mul_ii_1", "JIT_mul_i1_i1",

    "JIT_div_r_r", "JIT_div_r_r1", "JIT_div_r1_r", "JIT_div_r1_r1",
    "JIT_div_r_i", "JIT_div_r_i1", "JIT_div_r1_i", "JIT_div_r1_i1",
    "JIT_div_i_r", "JIT_div_i_r1", "JIT_div_i1_r", "JIT_div_i1_r1",
    "JIT_div_i_i", "JIT_div_i_i1", "JIT_div_ii_1", "JIT_div_i1_i1",

    "JIT_pow_r_r", "JIT_pow_r_r1", "JIT_pow_r1_r", "JIT_pow_r1_r1",
    "JIT_pow_r_i", "JIT_pow_r_i1", "JIT_pow_r1_i", "JIT_pow_r1_i1",
    "JIT_pow_i_r", "JIT_pow_i_r1", "JIT_pow_i1_r", "JIT_pow_i1_r1",
    "JIT_pow_i_i", "JIT_pow_i_i1", "JIT_pow_ii_1", "JIT_pow_i1_i1",

    "JIT_mod_r_r", "JIT_mod_r_r1", "JIT_mod_r1_r", "JIT_mod_r1_r1",
    "JIT_mod_r_i", "JIT_mod_r_i1", "JIT_mod_r1_i", "JIT_mod_r1_i1",
    "JIT_mod_i_r", "JIT_mod_i_r1", "JIT_mod_i1_r", "JIT_mod_i1_r1",
    "JIT_mod_i_i", "JIT_mod_i_i1", "JIT_mod_ii_1", "JIT_mod_i1_i1",

    "JIT_idiv_r_r", "JIT_idiv_r_r1", "JIT_idiv_r1_r", "JIT_idiv_r1_r1",
    "JIT_idiv_r_i", "JIT_idiv_r_i1", "JIT_idiv_r1_i", "JIT_idiv_r1_i1",
    "JIT_idiv_i_r", "JIT_idiv_i_r1", "JIT_idiv_i1_r", "JIT_idiv_i1_r1",
    "JIT_idiv_i_i", "JIT_idiv_i_i1", "JIT_idiv_ii_1", "JIT_idiv_i1_i1",

    "JIT_last"
};

static R_INLINE void genjit(JIT_OPCODE opcode, CSEXP operand, FUNC_TYPE f,
                            int type, int result_len, CSEXP resultTemplate,
                            CSEXP sym, CSEXP env);

static void unjit(Rboolean clearJittedBitsFlag);

//-----------------------------------------------------------------------------
// Printing routines

static const char *jitOpcodeName(JIT_OPCODE opcode)
{
    if ((int)opcode >= (int)JIT_endop && (int)opcode <= (int)JIT_last)
        return JIT_OPCODE_NAMES[opcode] + 4;    // +4 skips "JIT_"
    else
        return "[unknown JIT_OPCODE]";
}

static void printJitOp(const JIT_OP *op)        // print a JIT_OP
{
    CSEXP operand = op->operand;
    CSEXP result = op->result;

    printf("%-16s ", jitOpcodeName(op->opcode));

    if (result != R_NilValue)
        printf("result %s length %d ",
               type2char(TYPEOF(result)), LENGTH(result));

    if (operand == R_NilValue)
        printf("\n");
    else switch (TYPEOF(operand)) {
        case LGLSXP:
        case INTSXP:
        case REALSXP:
        case SYMSXP:
            printfSxp(operand, "operand");
            break;
        case LISTSXP:   // operand is a location
            printfSxp(getSymFromLoc(operand), "operand");
            break;
        default:
            printfSxp(operand, "illegal operand");
            UNIMPLEMENTED_TYPE("printJitOp", operand);
            break;
    }
}

static void printJitRecord(JIT_RECORD *prec)    // print a JIT_RECORD
{
    const JIT_OP *op = prec->ops;

    Dassert(NBR_ELEMS(JIT_OPCODE_NAMES) == (int)JIT_last + 1);
    Dassert(0 == strcmp(JIT_OPCODE_NAMES[JIT_last], "JIT_last"));

    printf("JIT type %s[%d] expression %s\n",
           type2char(TYPEOF(prec->ans)), LENGTH(prec->ans),
           deparseAsShortString(prec->original));

    do {
        printf("        ");
        printJitOp(op);
    } while ((op++)->opcode != JIT_endop);

    printf("\n");
}

void printJit(CSEXP s)                          // print a JITSXP
{
    assert(TYPEOF(s) == JITSXP);
    assert(TYPEOF(s->u.jitsxp.pjit) == RAWSXP);
    printJitRecord((JIT_RECORD *)RAW(s->u.jitsxp.pjit));
}

//-----------------------------------------------------------------------------
// Garbage collection routines

static R_INLINE void forwardJitOp(const JIT_OP *op, SEXP *pForwardedNodes)
{
     // "result" is the only field that is allocated by the jitter.
     // The other fields are protected by their original owners.

     if (op->result != R_NilValue)
         ForwardNode(op->result, pForwardedNodes);
}

static R_INLINE void forwardJitRecord(const SEXP pjit, int nops,
                                      SEXP *pForwardedNodes)
{
    int i;
    const JIT_OP *op;
    JIT_RECORD * const prec = (JIT_RECORD *)RAW(pjit);
    Dassert(TYPEOF(pjit) == RAWSXP);
    ForwardNode(pjit, pForwardedNodes);
    if (prec->original != R_NilValue)
        ForwardNode(prec->original, pForwardedNodes);
     if (prec->ans != R_NilValue)
         ForwardNode(prec->ans, pForwardedNodes);

    // if nops >= 0 then use it, else look for JIT_endop

    if (nops >= 0)
        for (i = 0; i < nops; i++)
            forwardJitOp(prec->ops + i, pForwardedNodes);
    else
        for (op = prec->ops; op->opcode != JIT_endop; op++)
            forwardJitOp(op, pForwardedNodes);
}

// Tell the garbage collector not to collect data in this file
// i.e. to not collect used elements in jitted[] or genex[].
//
// Elements of jitted[] from 0 to njitted-1 must be protected.
// Elements of genex.ops[] from 0 to ngenex-1 must be protected.
//
// We only need to protect fields that were allocated by the jitter
// because other SEXPs are protected by their original owners.
// All allocation for the jitter is done in this file, by allocVector
// and allocSExp and possibly others)

void ForwardJitNodes(SEXP *pForwardedNodes)
{
    int i;
    if (jitTrace >= 3)
        printf("#\t\t\t\tForwardJitNodes\n");

    for (i = 0; i < njitted; i++) {
        SEXP p = jitted[i];
        SEXP pjit = p->u.jitsxp.pjit;

        Dassert(TYPEOF(p) == JITSXP);
        Dassert(TYPEOF(pjit) == RAWSXP);

        ForwardNode(p, pForwardedNodes);        // RA_TODO necessary?
        forwardJitRecord(pjit, -1, pForwardedNodes);
    }
    if (genex)
        forwardJitRecord(genex, ngenex, pForwardedNodes);
}

//-----------------------------------------------------------------------------
// Routines for pushing and popping jit states.
// These are invoked as we enter and leave R functions.
// They are needed because jit(1) only applies in the func where it is used.
//
// For now for simplicity we don't keep a full state stack.  The
// disadvantage from the user's perspective is that he can't jit a
// function that is called from a function that is already being
// jitted --- do_jit() will issue a warning.
// In the long term this limitation should be lifted.

static int      istatestack;
static unsigned pushedJitDirective;
static int      pushedJitDepth;
static int      pushedJitGenDepth;
static int      pushedjitUnresolved;

static R_INLINE void pushJitStateAux(SEXP call)
{
    pushedJitDirective   = jitDirective;
    pushedJitDepth       = jitDepth;
    pushedJitGenDepth   = jitGenDepth;
    pushedjitUnresolved = jitUnresolved;
    jitDirective     = 0;
    jitDepth         = 0;
    jitGenDepth      = 0;
    jitUnresolved = 0;
}

void pushJitState(SEXP call)    // called on entry to an R function
{
    if (globalJitDirective) {
        if (jitTrace >= 3)
            printf("#\t\t\t\tpushJitState istatestack %d expression %s\n",
                   istatestack, deparseAsShortString(call));
        if (istatestack++ == 0)
            pushJitStateAux(call);
    }
}

void popJitState(SEXP call)     // called on exit from an R function
{
    if (globalJitDirective) {
        Dassert(istatestack >= 0);
        if (istatestack == 0) {
            if (jitTrace >= 3)
                printfSxp(call, "# Return from jitted function");
            jitOff(FALSE);
        } else if (--istatestack == 0) {
            jitDirective     = pushedJitDirective;
            jitDepth         = pushedJitDepth;
            jitGenDepth      = pushedJitGenDepth;
            jitUnresolved = pushedjitUnresolved;
        }
        if (jitTrace >= 3)
            printf("#\t\t\t\tpopJitState istatestack %d expression %s\n",
                   istatestack, deparseAsShortString(call));
    }
}

//-----------------------------------------------------------------------------
// Return s->original, needed for deparse

SEXP getExpBeforeItWasJitted(CSEXP s)
{
    Dassert(TYPEOF(s) == JITSXP);
    Dassert(TYPEOF(CAR(s)) == RAWSXP);
    return ((JIT_RECORD *)RAW(CAR(s)))->original;
}

// Get the name of R function that called the R function "jit"
// i.e. return the name of the caller's caller,
// For nice err messages.

static const char *getNameOfFuncThatCalledJit(void)
{
    const char *str = NULL;
    RCNTXT *cptr;
    int nesting = 0;
    for (cptr = R_GlobalContext;
         cptr != NULL && cptr->callflag != CTXT_TOPLEVEL;
         cptr = cptr->nextcontext)
        if (cptr->callflag & (CTXT_BROWSER | CTXT_FUNCTION)) {
            if (nesting++ == 1) { // must get caller's caller
                SEXP fun = CAR(cptr->call);
                if (TYPEOF(fun) == SYMSXP)
                    str = CHAR(PRINTNAME(fun));
                break;
            }
        }
    return str;
}

static void nestedJitWarning(const unsigned jit, const unsigned trace)
{
    char thisFunc[STRLEN+1], originalFunc[STRLEN+1];

    const char *str = getNameOfFuncThatCalledJit();
    thisFunc[0] = 0;
    if (str)
        snprintf(thisFunc, STRLEN, _("in %s "), str);

    if (jitFuncName)
        snprintf(originalFunc, STRLEN, "%s ", jitFuncName);
    else
        snprintf(originalFunc, STRLEN, _("a calling function "));

    warning(_("ignored jit(%d,%d) %sbecause %sis already jitting"),
            jit, trace, thisFunc, originalFunc);
}

static SEXP getJitReturnVal(void)       // just a little helper for do_jit
{
    SEXP ans;
    PROTECT(ans = allocVector(INTSXP, 4));
    INTEGER(ans)[0] = (globalJitDirective & JIT_ON) != 0;
    INTEGER(ans)[1] = jitTrace;
    INTEGER(ans)[2] = (globalJitDirective & JIT_RANGE_CHECK) != 0;
    INTEGER(ans)[3] = (istatestack == 1) && (pushedJitDirective & JIT_ON);
    UNPROTECT(1);
    return ans;
}

// called by do_jit after initing globalJitDirective and jitTrace

static void doJitAux(const SEXP call)
{
    if (globalJitDirective == 0)
        unjit(TRUE);
    jitDirective = globalJitDirective;
    jitInit();
    if (globalJitDirective == 0)
        jitDepth = 0;

    // stash caller's name in jitFuncName (for nice error reporting)

    jitFuncName = NULL;
    if (globalJitDirective)
        jitFuncName = getNameOfFuncThatCalledJit();

    // following is needed to counteract the popping effect of the
    // return from the R function "jit" which called "do_jit"

    istatestack = 1;
    pushJitStateAux(call);
 }

// User callable R function.
// This changes or reads globalJitDirective and jitTrace, and
// calls doJitAux to change other global jit variables.
//
// Usage: jit(jit=NA, trace=0)
//
// Returns a 4 element int vector showing the setting for jit, trace,
// the JIT_RANGE_CHECK bit setting (currently always 1), and
// the jitDirective setting for the current function.

SEXP attribute_hidden do_jit(SEXP call, SEXP op, SEXP args, SEXP env)
{
    int jit, trace;

    checkArity(op, args);

    jit = asInteger(CAR(args));
    if (jit != NA_INTEGER && jit != 0 && jit != 1)
        error(_("\"jit\" parameter is %d, legal values are NA, 0, or 1"), jit);

    trace = asInteger(CADR(args));
    if (trace < 0 || trace > 3) {
        if (trace == NA_INTEGER)
            error(_("\"trace\" parameter is NA, legal values are 0 to 3"));
        else
            error(_("\"trace\" parameter is %d, legal values are 0 to 3"),
                  trace);
    }
    if (jit != NA_INTEGER) {
        if (trace >= 3 || jitTrace >= 3)
            printf("# jit(%d, %d) invoked by %s\n",
                   jit, trace, getNameOfFuncThatCalledJit());

        if (trace)
            jitTrace = trace;   // set tracing now, so trace doJitAux

        if (globalJitDirective && istatestack > 1) {
            // warn if user tried to change jit state in a
            // function called from a jit block

            if (jitDirective != jit)
                nestedJitWarning(jit, trace);

        } else {
            // following "if" prevents state changes if already jitting

            if (!(globalJitDirective && jit)) {
                globalJitDirective = jit? (JIT_ON | JIT_RANGE_CHECK) : 0;
                doJitAux(call);
            }
        }
        jitTrace = jit? trace: 0;
    }
    return getJitReturnVal();
}

//-----------------------------------------------------------------------------
// Routines which make sure that the jitted environment is stable.
//
// The rules are (a) the user cannot change the binding type or length of a
// jitted variable, (b) the user cannot unbind or remove a jitted variable.
// If the user tries to violate these rules we issue an error message.
//
// These rules are needed because
// (a) we store pointers directly to bindings in the "operand" field
//     of jit instructions
// (b) jitted opcodes assume fixed types e.g. add_r1_r1 assumes
//     real scalar operands
// (c) we preallocate fixed length vectors for jit instruction operands.
//
// The rules apply not only to the function with the jit block
// but also to all functions called from the jit block.
//
// We keep track of which bindings are jitted with the gp JITTED_BIT.


// jitError is like error but appends "jitting was enabled in func" msg

static void jitError(const char *format, ...)
{
    #define BUFSIZE 200
    char buf[BUFSIZE];
    int nchars;
    va_list(ap);
    assert(BUFSIZE < R_WarnLength);
    va_start(ap, format);
    nchars = vsnprintf(buf, BUFSIZE-1, format, ap);
    assert(nchars >= 0);
    buf[nchars] = 0;    // guarantee null termination
    va_end(ap);
    // we append a msg if jitting was enabled in a func that called this func
    if (jitFuncName && istatestack > 0)
        error("%s (jitting was enabled in \"%s\")", buf, jitFuncName);
    else
        error(buf);
}

static R_INLINE void setJittedBit(SEXP location)
{
    location->sxpinfo.gp |= JITTED_BIT;
}

static R_INLINE void clearJittedBit(SEXP location)
{
    if (jitTrace >= 3 && (location->sxpinfo.gp & JITTED_BIT))
        printfSxp(location, "#   Clear jitted bit");

    location->sxpinfo.gp &= ~JITTED_BIT;
}

static R_INLINE int isJitted(SEXP b)    // strictly, "is POSSIBLY jitted"
{
    return b->sxpinfo.gp & JITTED_BIT;
}

// This is called when we are about to change a binding value.  If the
// binding is jitted allow the new binding only if is the same length
// and type as the current binding.
//
// The "val != R_NilValue" check below is needed because routines like
// do_for() first redefine the loop var with defineVar(sym, R_NilValue, env)
// and then redefine it again for each iteration with setVar(sym, v, env).
// When the the loop var is jitted, we allow the first call,
// with R_NilValue, even though it is changing the type of a jitted sym.
// So with this code there is a risk that some R code elsewhere will set a
// jitted var to R_NilValue and not set it back or set it to another type
// before evalJit() is invoked.  In practice that does not happen.  With
// DEBUG_JIT > 0, the Dasserts in evalJIT will catch it if it ever does.

void checkJitBindingVal(const SEXP b, const SEXP val)
{
    if (isJitted(b)) {
        SEXP oldval = CAR(b);
        const SEXP sym = TAG(b);
        Dassert(TYPEOF(b) == LISTSXP);
        Dassert(TYPEOF(sym) == SYMSXP);

        if (TYPEOF(oldval) == PROMSXP)
            oldval = PRVALUE(oldval);

        if (val != R_NilValue &&        // see comment in above header
            oldval != R_NilValue) {     // not initing or set to NIL last time?

            if (TYPEOF(oldval) != TYPEOF(val))
                jitError(_("cannot change the type of a jitted variable\n"
                            "Tried to change \"%s\" from %s to %s"),
                         CHAR(PRINTNAME(sym)),
                         type2char(TYPEOF(oldval)),
                         type2char(TYPEOF(val)));

            if (LENGTH(oldval) != LENGTH(val))
                jitError(_("cannot change the length of a jitted variable\n"
                           "Tried to change \"%s\" from length %d to length %d"),
                         CHAR(PRINTNAME(sym)), LENGTH(oldval), LENGTH(val));
        }
    }
}

// Like checkJitBindingVal above but called when we bind the
// value directly to the symbol rather than through a frame.

void checkJitSymBindingVal(const SEXP sym, const SEXP val)
{
    if (isJitted(sym) && sym != R_TmpvalSymbol) {
        const SEXP oldval = CDR(sym);   // SYMSXP so CDR is value field
        Dassert(TYPEOF(sym) == SYMSXP);
        Dassert(TYPEOF(oldval) != PROMSXP);

        if (TYPEOF(oldval) != TYPEOF(val))
            jitError(_("cannot change the type of jitted variable\n"
                       "Tried to change \"%s\" from %s to %s"),
                     CHAR(PRINTNAME(sym)),
                     type2char(TYPEOF(oldval)),
                     type2char(TYPEOF(val)));

        if (LENGTH(oldval) != LENGTH(val))
            jitError(_("cannot change the length of a jitted variable\n"
                       "Tried to change \"%s\" from length %d to length %d"),
                     CHAR(PRINTNAME(sym)), LENGTH(oldval), LENGTH(val));
    }
}

// This func is called by defineVar.  It is only be called if sym
// is not in the current frame.
//
// This func handles the case shown in the example R function below.
// The essential characteristic of the exampleis that it first uses
// a non-local variable and then assigns to the non-local variable (thereby
// creating a local copy).
//
//  glob = 1
//  foo <- function()
//  {
//      jit(1, 1)
//      for (i in 1:3) {    # need a loop for jitting to take place
//          x <- glob + 1   # line 7, after this glob is marked as jitted
//          glob <- 1:3     # line 8, this changes glob to local scope
//     }
//  }
//
// On the first iter through the loop we use the global glob for
// x <- glob + 1 (line 7). The following assignment to glob (line 8)
// creates a local glob. On the next iteration through the loop, the jitter
// incorrectly still points to the global glob, instead of the local
// glob.  To prevent this, the function below disallows the assignment
// on line 8.
//
// This function slows down assignment to non local vars in jit blocks
// (because we have search beyond the local frame with findVarLoc).
// The slowdown goes away if the assignment is jitted.

//-----------------------------------------------------------------------------
void checkLocalBindingForJit(const SEXP sym, const SEXP val, const SEXP env)
{
    if (jitDirective) {
        SEXP loc = findVarLoc(sym, env);
        if (isJitted(loc)) {
            const char *name = CHAR(PRINTNAME(getSymFromLoc(loc)));
            jitError(
                _("Cannot change jitted symbol \"%s\" to local scope\n"
                  "Possible remedy: make \"%s\" local by "
                  "assigning to it before the loop"),
                name, name);
        }
    }
}

void checkJitRemove(CSEXP loc)
{
    if (isJitted(loc))
       jitError(_("cannot remove jitted variable \"%s\""),
                CHAR(PRINTNAME(getSymFromLoc(loc))));
}

// This prevents the user doing anything that could affect stability of
// the jitting environment. The rule is: if in doubt, use it.

void disallowIfJitting(const char s[])
{
    if (globalJitDirective)
        jitError(_("cannot %s when jitting"), s);
}

// This could be removed, but for now and for paranoia check that
// findVar(e, env) == getSymValFromLoc(findVarLoc(e, env)).
//
// Note that this is called by evalSYM() even when the jitter is
// not running i.e. it slows down standard R evaluation.

void checkLocationConsistency(SEXP e, SEXP env, SEXP loc, SEXP val)
{
#if DEBUG_SYM
    SEXP findVarVal = findVar(e, env);
    assert(loc != R_NilValue);
    if (jitTrace >= 3 && traceEvalFlag)
        printf("checkLocationConsistency\n");
    if (val != findVarVal) {
        // must print before calling assertFail because error itself
        // can invoke this function via a call to eval()

        REprintf("Internal error in checkLocationConsistency: "
                 "findVarVal 0x%x != val 0x%x loc 0x%x\n",
                 findVarVal, val, loc);
        REprintf("findVar val: "); printSxp(findVarVal, env, FALSE);
        REprintf("val: "); printSxp(val, env, FALSE);
        REprintf("loc: "); printSxp(loc, env, FALSE);

        assertFail(__FILE__, __LINE__, "see above messages");
    }
#endif
}

// these trace functions are for debugging and could disappear eventually

void traceBinding(const SEXP b, const SEXP val)
{
     if (DEBUG_JIT >= 3 && globalJitDirective) {
        printfSxp(TAG(b), "\nSET_BINDING_VALUE");
        printfSxp(CAR(b), "  current");
        printfSxp(val, "  new    ");
     }
}

void traceSymBinding(const SEXP sym, const SEXP val)
{
     if (DEBUG_JIT >= 3 && globalJitDirective) {
        Dassert(TYPEOF(sym) == SYMSXP);
        printfSxp(sym, "\nSET_SYMBOL_BINDING_VALUE current");
        printfSxp(val, "                         new    ");
    }
}

//-----------------------------------------------------------------------------
// Routines which control the JIT compiler's state

void decJitUnresolved(void)
{
    if (jitCompiling()) {
        jitUnresolved--;
        Dassert(jitUnresolved >= 0);
    }
}

static R_INLINE void assertSxpInfoZero(SEXP s) // play safe
{
    Dassert(TYPEOF(s) == LANGSXP);
    assert(s->sxpinfo.obj == 0);
    assert(s->sxpinfo.named == 0);
    assert(s->sxpinfo.gp == 0);
    // assert(s->sxpinfo.mark == 0);
    if (s->sxpinfo.debug)
        disallowIfJitting("use \"debug\"");
    if (s->sxpinfo.trace)
        disallowIfJitting("use \"trace\"");
    assert(s->sxpinfo.spare == 0);
    // assert(p->sxpinfo.gcgen == 0);
    assert(s->sxpinfo.gccls == 0);
}

// return TRUE if op->operand is a pointer to a location i.e. a binding

static R_INLINE Rboolean isOperandALocation(JIT_OPCODE opcode)
{
    switch (opcode) {
        case JIT_pushsym_r:
        case JIT_pushsym_i:
        case JIT_Cpushsym_r:
        case JIT_Cpushsym_i:
        case JIT_assign_r_r:
        case JIT_assign_r_r1:
        case JIT_assign_r1_r:
        case JIT_assign_r1_r1:
        case JIT_assign_r_i:
        case JIT_assign_r_i1:
        case JIT_assign_r1_i:
        case JIT_assign_r1_i1:
        case JIT_assign_i_r:
        case JIT_assign_i_r1:
        case JIT_assign_i1_r:
        case JIT_assign_i1_r1:
        case JIT_assign_i_i:
        case JIT_assign_i_i1:
        case JIT_assign_i1_i:
        case JIT_assign_i1_i1:
            return TRUE;
        default:
            break;
    }
    return FALSE;
}

// Clear the JITTED_BIT for all symbols referenced by the given JIT_RECORD prec.
// NB the environment when the jit record was created must still be in place
// because we access binding locations by direct pointers.

static R_INLINE void clearJittedBits(JIT_RECORD *prec)
{
    const JIT_OP *op;
    for (op = prec->ops; op->opcode != JIT_endop; op++) {
        if (isOperandALocation(op->opcode))
            clearJittedBit(op->operand);
    }
}

// Convert jitted expressions in jitted array back to their unjitted form.
// Also clear all jitted symbols' JITTED_BIT.
// Called by jit(0) and when returning from a function that invoked jit(1).

static void unjit(Rboolean clearJittedBitsFlag)
{
    if (njitted) { // pre-test for efficiency
        int i;
        for (i = 0; i < njitted; i++) {
            int gcgen;
            SEXP p = jitted[i];
            SEXP pjit = p->u.jitsxp.pjit;
            JIT_RECORD *prec = (JIT_RECORD *)RAW(pjit);

            Dassert(TYPEOF(p) == JITSXP);
            Dassert(TYPEOF(pjit) == RAWSXP);

            if (jitTrace >= 3)
                printfSxp(p, "# Unjit", p);

            // restore original expression
            assertSxpInfoZero(prec->original);
            gcgen = p->sxpinfo.gcgen;
            p->sxpinfo = prec->original->sxpinfo;
            p->sxpinfo.gcgen = gcgen;   // RA_TODO revisit!
            ATTRIB(p) = ATTRIB(prec->original);
            p->u.listsxp = prec->original->u.listsxp;  // update car, cdr, tag

            if (clearJittedBitsFlag)
                clearJittedBits(prec);
        }
        njitted = 0;
    }
}

// Init the just-in-time generation code for the current user eval().
// Called from the main read-eval-print loop and by do_jit(0).

void jitInit(void)
{
    if (jitDirective) {
        jitDepth = 1;
        jitGenDepth = 0;
        genex = NULL;
        ngenex = 0;
        jitCompileAborted = FALSE;
    }
}

static R_INLINE void jitOffWarning(Rboolean issueWarning)
{
    // Using REprintf is not the best way to issue a warning,
    // but we don't want to call warning() while in error (this
    // function can be invoked from error).  The test against
    // jitFuncName ensures that this msg is only issued if the
    // user is jitting at the prompt level.

    if (issueWarning && globalJitDirective && jitFuncName == NULL)
        REprintf("\nWarning: jitting is now disabled\n");
}

// turn off jitting -- called on return from a jitted function or on error

void jitOff(Rboolean issueWarning)
{
    jitOffWarning(issueWarning);
    unjit(TRUE);
    globalJitDirective = 0;
    jitDirective = 0;
    jitDepth = 0;
    jitGenDepth = 0;
    jitTrace = 0;
    genex = NULL;
    ngenex = 0;
    jitCompileAborted = FALSE;
    istatestack = 0;
    printSxpDepth = 0;  // not the cleanest place for this, but convenient
}

static void compiledMessage(CSEXP s, JIT_RECORD *prec)
{
    assert(TYPEOF(s) == JITSXP);

    if (jitTrace >= 1) {
        printf(_("#   Compiled %s\n"),
               deparseAsShortString(prec->original));

        if (jitTrace >= 3) {
            printf("#\tto ");
            printJit(s);
        }
     }
}

// Called on successful jit generation i.e. we reached the end of the
// expression e being compiled, and were able to compile
// everything in the expression.
// This changes e to a JITSXP. See jit-data-structures*.jpg.

static void goodCompile(SEXP e, CSEXP resultTemplate, CSEXP op)
{
    JIT_RECORD *prec = (JIT_RECORD *)RAW(genex);
    SEXP ans, original;
    Dassert(ngenex > 0);
    Dassert(LENGTH(resultTemplate) > 0);
    Dassert(jitUnresolved == 0);
    jitUnresolved = 1;  // for check in decJitUnresolved in genjit
    genjit(JIT_endop, R_NilValue, NULL,
           NILSXP, 0, R_NilValue, R_NilValue, R_NilValue);

    // allocate both ans and original before we use them, in case alloc causes a gc
    PROTECT(ans = allocVector(TYPEOF(resultTemplate), LENGTH(resultTemplate)));
    PROTECT(original = allocSExp(TYPEOF(e)));
    prec->ans = ans;
    assertSxpInfoZero(e);
    prec->original = original;                  // save original expression
    prec->original->u.listsxp = e->u.listsxp;   // set car,cdr, and tag
    ATTRIB(prec->original) = ATTRIB(e);
    SET_TYPEOF(e, JITSXP);
    SETCAR(e, genex);                           // car is pjit
    SETCDR(e, R_NilValue);                      // not strictly necessary
    SET_TAG(e, R_NilValue);                     // not strictly necessary
    assert(njitted < NBR_ELEMS(jitted));
    jitted[njitted++] = e;
    UNPROTECT(2);
    compiledMessage(e, prec);
}

// Unsuccessful just-in-time generation i.e. we reached the end of the
// expression e being compiled, and couldn't compile everything
// in the expression.
//
// Prevent further attempts to compile e by setting CANNOT_JIT_BIT.
// Note that sub-expressions of e could still be jittable.

static R_INLINE void markAsNotJittable(SEXP e)
{
    e->sxpinfo.gp |= CANNOT_JIT_BIT;

    if (jitTrace >= 2)
        printf(_("#   Did not compile %s\n"), deparseAsShortString(e));
}

// Possibly start the just-in-time compiler on e, where e is the LANGSXP
// expression currently being evaluated by eval().
// Once we have started the compiler, R code that is executed after
// this will check when jitGenDepth is non-zero and if so append
// jit ops into genex by calling genjit().

void jitProlog(CSEXP e, CSEXP op, const char msg[])
{
    Rboolean fireUp = !(e->sxpinfo.gp & CANNOT_JIT_BIT) &&
                      jitGenDepth++ == 0;       // note ++

    if (jitTrace >= 3)
        printf("#\t\t\t\t%d jitProlog invoked by \"%s\", %s%s %s\n",
               R_EvalDepth, msg, (fireUp? "fire up compiler for ": ""),
               ((e->sxpinfo.gp & CANNOT_JIT_BIT)?
                    "not jittable": "possibly jittable"),
               deparseAsShortString(e));

    if (fireUp) {
        Dassert(TYPEOF(e) == LANGSXP);
        Dassert(TYPEOF(CAR(e)) == SYMSXP);
        Dassert(globalJitDirective);
        Dassert(jitDirective);
        Dassert(jitDepth >= 2);

        // Fire up the jit generation machine.
        // Allocate a JIT_RECORD to store the generated jit instructions.
        // We allocate genex as an RAWSXP but use it as a JIT_RECORD.
        // It gets protected from garbage collection in ForwardJitNodes.

        JIT_RECORD *prec;
        genex = allocVector(RAWSXP, sizeof(JIT_RECORD));
        prec = (JIT_RECORD *)RAW(genex);
        prec->original = R_NilValue;
        prec->ans = R_NilValue;
        ngenex = 0;
        jitUnresolved = 1;
        pendingPromisesAtJitProlog = R_PendingPromises;
    }
}

void jitEpilog(SEXP e, CSEXP resultTemplate, CSEXP op)
{
    if (jitTrace >= 3) {
        printf("#\t\t\t\t%d jitEpilog ", R_EvalDepth);
        if (e->sxpinfo.gp & CANNOT_JIT_BIT)
            printf("not jittable ");
        else
            printf("jitGenDepth %d done ", jitGenDepth);
        printf("%s\n", deparseAsShortString(e));
    }
    if (jitGenDepth) {
        Dassert(jitDirective);
        Dassert(jitDepth >= 2);
        Dassert(jitGenDepth >= 0);
        Dassert(TYPEOF(e) == LANGSXP);

        if (--jitGenDepth == 0) {               // reached end of expression?
            if (jitCompileAborted) {
                markAsNotJittable(e);
                jitCompileAborted = FALSE;     // RA_TODO unneeded?
            } else if (jitUnresolved) {
                // Nothing was generated or not everything in e was compiled.
                // Mark e as not jittable, unless the reason we didn't
                // compile everything in the expression was because we were
                // evaluating a promise --- so we can try again on next iter.

                if (R_PendingPromises == pendingPromisesAtJitProlog)
                    markAsNotJittable(e);

            } else if (njitted >= NBR_ELEMS(jitted))
                markAsNotJittable(e);
            else
                goodCompile(e, resultTemplate, op);
        }
    }
}

// This is called during a compilation if we know we should not continue
// e.g. we reached the end of a basic block of R code.
// Used as efficiency measure too, if there is no point in continuing
// within a basic block.
//
// RA_TODO would be more efficient if we did a jitOff here so we don't
// keep on invoking genjit (which simply returns).  But that involves
// tricky flag manipulation.

void cannotJitCompile(CSEXP e)
{
    if (jitGenDepth) {
        if (jitTrace >= 3 && !jitCompileAborted)
            printf("#       Aborted compile\t\t%d "
                   "Mark as not compilable because reached %s\n",
                   R_EvalDepth, deparseAsShortString(e));

        jitCompileAborted = TRUE;
    }
}

void jitEnterLoop(CSEXP s)      // invoked on entry to an R loop
{
    // Abort current jit compilation, if any, because the "for" statement
    // (and friends) is not compilable itself (although the code IN the
    // loop may be).  This prevents the jitter from compiling repeated
    // iterations of the loop as one huge sequence of jitops.

    cannotJitCompile(s);

    // We call deparseAsShortString BEFORE bumping jitDepth, else
    // deparseAsShortString itself could call eval which is confusing.

    if (jitDepth == 1 && jitTrace >= 2)
        printf(_("# Begin JIT compilation: %s\n"), deparseAsShortString(s));

    if (jitDepth > 0)
        jitDepth++;
}

void jitExitLoop(CSEXP s)       // invoked on exit from an R loop
{
    if (jitDepth > 1)
        jitDepth--;

    if (jitDepth == 1 && jitTrace >= 2)
        printf(_("# End JIT compilation: %s\n\n"), deparseAsShortString(s));
}

// Get the c function associated with the LANGSXP s if possible.
// If not possible, return NULL.

CCODE getCfun(CSEXP s)
{
    CCODE cfun = NULL;
    if (TYPEOF(s) == LANGSXP) {
        const SEXP car = CAR(s);
        if (TYPEOF(car) == SYMSXP) {
                const SEXP symvalue = SYMVALUE(car);
                if (TYPEOF(symvalue) == SPECIALSXP)
                    cfun = PRIMFUN(symvalue);
        }
    }
    return cfun;
}

// Return true if e is a subassignment expression.  Like this: x[i] <- rhs
//
// This is for efficiency so we don't keep on retrying to jit
// the subscripted assignment every time through the loop.
// The usual technique of setting the CANNOT_JIT_BIT doesn't work here
// because eval() creates a new expression for "x[i]<- rhs" every time
// through the loop.
//
// This is something of a catch 22 because this function itself is
// slow.  But this function will dissapear once we add code
// to jit `[<-`.

static R_INLINE Rboolean isSubassign(SEXP e)
{
    SEXP rhs = CADR(e);
    CCODE cfun = getCfun(rhs);
    return (cfun == do_subassign)  ||
           (cfun == do_subassign2) ||
           (cfun == do_subassign3);
}

// strictly, this returns "is assignment POTENTIALLY jittable"

Rboolean isAssignmentJittable(SEXP rhs)
{
    return (jitDepth >= 2) && !jitCompileAborted && !isSubassign(rhs);
}

//-----------------------------------------------------------------------------
// Routines to generate JIT instructions

static void printGenjitOp(JIT_OP *op)   // a helper for genjitAux
{
    if (jitTrace >= 3) {
        printf("#\tGENERATE ");
        printJitOp(op);
        printf("\n");
    }
}

// Generate the given jit instruction i.e. append the instruction to
// the jit instructions in genex.
//
// A note on NAMED handling: all fields of a JIT_RECORD are read only
// except the ones we create ourselves (with allocVector and friends).
// So we don't need to check NAMED.

static void genjitAux(JIT_OPCODE opcode, CSEXP operand, FUNC_TYPE func,
                      int resultType, int resultLen, CSEXP resultTemplate,
                      CSEXP sym, CSEXP env)
{
    JIT_RECORD * const prec = (JIT_RECORD *)RAW(genex);
    JIT_OP * const op = prec->ops + ngenex;

    Dassert(ngenex < NBR_ELEMS(prec->ops));
    Dassert(jitDirective);
    Dassert(jitDepth >= 2);
    Dassert(jitGenDepth > 0 || opcode == JIT_endop);
    Dassert(opcode >= JIT_endop && opcode < JIT_last);

    ngenex++;
    op->opcode = opcode;
    op->operand = operand;
    op->func = func;
#if DEBUG_SYM
    op->sym = sym;
    op->env = env;
#endif
    op->result = R_NilValue; // preassign (for ForwardJitNodes,
                             // because allocVector may cause gc)
    if (resultLen == 0 && resultTemplate != R_NilValue)
        resultLen = LENGTH(resultTemplate);
    if (resultLen) {
        // PROTECT here because this can be called from integer_binary and
        // friends which don't protect "ans" i.e. "resultTemplate".
        PROTECT(resultTemplate);
        if (resultType == NILSXP)
            resultType = TYPEOF(resultTemplate);
        op->result = allocVector(resultType, resultLen);
        UNPROTECT(1);
    }
    decJitUnresolved();
    printGenjitOp(op);
}

static void genjit(JIT_OPCODE opcode, CSEXP operand, FUNC_TYPE func,
                   int resultType, int resultLen, CSEXP resultTemplate,
                   CSEXP sym, CSEXP env)
{
    if (jitCompileAborted) {
        if (jitTrace >= 3)
            printf("#\t\t\t\t%d Skipped generate %s because "
                   "jitCompileAborted==TRUE\n",
                   R_EvalDepth, jitOpcodeName(opcode));

    } else if (ngenex >= MAX_JIT_EXP_LEN) {
        // the expression is too long to be jitted.

        if (jitTrace >= 2)
            printf(_("#   Expression is too long to be jitted\n"));
        cannotJitCompile(R_NilValue);

    } else if (R_PendingPromises != pendingPromisesAtJitProlog) {
        // no jit generation while evaluating a promise

        if (jitTrace >= 3)
            printf("#\t\t\t\t%d Skipped generate %s because "
                   "currently evaluating a promise\n",
                   R_EvalDepth, jitOpcodeName(opcode));
    } else {
            // yay! we can finally generate the instruction

            genjitAux(opcode, operand, func,
                       resultType, resultLen, resultTemplate, sym, env);
    }
}

// change opcode by adding appropriate offset if we
// are generating range-checked code

static R_INLINE JIT_OPCODE adjustForRangeCheck(JIT_OPCODE opcode, int offset)
{
    if (jitDirective & JIT_RANGE_CHECK)
        return (JIT_OPCODE)((int)opcode + offset);  // used range-checking
    else
        return opcode;
}

// change opcode by adding offset for the types of operands x and y

static R_INLINE int adjustForTypes(JIT_OPCODE opcode, CSEXP x, CSEXP y)
{
    int adjust = 0;

    if (TYPEOF(x) == INTSXP || TYPEOF(x) == LGLSXP)
        adjust += (int)JIT_add_i_r - (int)JIT_add_r_r; // op_r_r becomes op_i_r

    if (TYPEOF(y) == INTSXP || TYPEOF(y) == LGLSXP)
        adjust += (int)JIT_add_r_i - (int)JIT_add_r_r; // op_r_r becomes op_r_i

    return (JIT_OPCODE)((int)opcode + adjust);
}

void genjitPush(JIT_OPCODE opcode, SEXP e)      // invoked for numeric constants
{
    const int offset = (int)JIT_Cpush_r - (int)JIT_push_r;
    jitUnresolved++;
    genjit(adjustForRangeCheck(opcode, offset), e, NULL,
           NILSXP, 0, R_NilValue, R_NilValue, R_NilValue);
}

void genjitPushsym(SEXP location, SEXP env)     // invoked for symbols
{
    SEXP sym;

    if (location == NULL)
        disallowIfJitting("use a \"...\" variable");

    Dassert(TYPEOF(location) == LISTSXP || TYPEOF(location) == SYMSXP);
    sym = getSymFromLoc(location);
    Dassert(TYPEOF(sym) == SYMSXP);

    if (sym != R_TmpvalSymbol) {
        JIT_OPCODE opcode = JIT_endop;
        const int offset = (int)JIT_Cpushsym_r - (int)JIT_pushsym_r;
        SEXP val = getSymValFromLoc(location);
        if (TYPEOF(val) == PROMSXP)
            val = PRVALUE(val);
        Dassert(val != R_UnboundValue && val != R_NilValue);
        switch(TYPEOF(val)) {
            case LGLSXP:    // treat logicals as integers
            case INTSXP:
                opcode = JIT_pushsym_i;
                break;
            case REALSXP:
                opcode = JIT_pushsym_r;
                break;
            default:
                break;
        }
        if (opcode != JIT_endop) {
            setJittedBit(location);
            genjit(adjustForRangeCheck(opcode, offset), location, NULL,
                   NILSXP, 0, R_NilValue, sym, env);
        }
    }
}

void genjitSubset(CSEXP x, CSEXP subset)        // for expressions like x[i]
{
    static const JIT_OPCODE subsetOps[4][2] =
    {
        { JIT_subset_r_r,  JIT_subset_r_r1 },
        { JIT_subset_r_i,  JIT_subset_r_i1 },
        { JIT_subset_i_r,  JIT_subset_i_r1 },
        { JIT_subset_i_i,  JIT_subset_i_i1 },
    };
    // RA_TODO this could be made more efficient
    if ((TYPEOF(x) == INTSXP || TYPEOF(x) == REALSXP) &&
            (TYPEOF(subset) == INTSXP || TYPEOF(subset) == REALSXP)) {
        const int icol = (LENGTH(subset) == 1);
        const int irow = 2 * (TYPEOF(x) == INTSXP) + (TYPEOF(subset) == INTSXP);
        JIT_OPCODE opcode = subsetOps[irow][icol];
        if (icol == 1) {  // RA_TODO this is just for now because evalJIT only
                          // supports subscripts of len 1
            const int offset = (int)JIT_Csubset_r_r - (int)JIT_subset_r_r;
            genjit(adjustForRangeCheck(opcode, offset), R_NilValue, NULL,
                   NILSXP, LENGTH(subset), x, R_NilValue, R_NilValue);
        }
    }
}

void DCheckGenjitAssignParams(CSEXP sym, SEXP location,
                              CSEXP value, CSEXP env, const char msg[])
{
#if DEBUG_JIT
    if (sym != location) {
        SEXP findVarLoc1 = findVarLoc(sym, env);
        assert(findVarLoc1 != R_UnboundValue);
        assert(getSymFromLoc(findVarLoc1) == sym);
        if (location != R_NilValue && location != findVarLoc1) {
            printf("Passed in from %s\n", msg);
            printfSxp(location,  "  location");
            printf("  sym "); printSxp(sym, env, TRUE);
            printfSxp(env, "  env");
            printfSxp(findVarLoc1, "Calculated\n  findVarLoc1");
            assert(location == findVarLoc1);    // will always fail
       }
    }
#endif
}

// Invoked for R expressions like sym = value.
// We have already generated code to push the value onto the stack
// We now generate a JIT_assign opcode with sym as the
// operand (sym is not pushed).

void genjitAssign(CSEXP sym, SEXP location,
                  CSEXP value, CSEXP env, const char msg[])
{
    SEXP symval;
    int ny = LENGTH(value);

    if (ny == 0 || // don't jit this weird case with a zero length value
            (TYPEOF(value) != LGLSXP && TYPEOF(value) != INTSXP &&
             TYPEOF(value) != REALSXP)) {
        return;
    }
    Dassert(TYPEOF(sym) == SYMSXP);
    Dassert(TYPEOF(location) == LISTSXP || TYPEOF(location) == SYMSXP);
    Dassert(location != R_NilValue && location != R_UnboundValue);

    DCheckGenjitAssignParams(sym, location, value, env, msg);
    symval = getSymValFromLoc(location);
    if (symval == R_UnboundValue) {
#if 0 // removed because causes later problem in evalJIT

        // Unusual special case for when the user creates a variable in a
        // loop in the R_GlobalEnv:
        //    jit(1,1)
        //    for(i in 1:3)
        //       x <- 1 + 2 # create and assign to new variable x
        // The symval will be unbound.
        // Note that this is not a problem if the above loop is in an R func.

        JIT_OPCODE opcode = ...;
        genjit(opcode, location, NULL, 0, 0, R_NilValue, sym, env);
#endif
    } else {
        JIT_OPCODE opcode = adjustForTypes(JIT_assign_r_r, symval, value);
        if (ny == 1)
            opcode += 3;                // r_r + 3 is r1_r1
        assert(ny == LENGTH(symval));   // check that both side have same len
        setJittedBit(location);
        genjit(opcode, location, NULL, 0, 0, R_NilValue, sym, env);
   }
}

static R_INLINE void intHelper(JIT_OPCODE opcode, CSEXP resultTemplate)
{
    genjit(opcode, R_NilValue, NULL,
           NILSXP, 0, resultTemplate, R_NilValue, R_NilValue);
}

// invoked via the inline function genjitIntBin for int arith ops like "+"

void genjitIntBinAux(JIT_OPCODE opcode, CSEXP x, CSEXP y,
                        CSEXP resultTemplate)
{
    const int nx = LENGTH(x);
    const int ny = LENGTH(y);

    if (ny == 1) {
        if (nx == 1)
            intHelper(opcode+3, resultTemplate);  // i_i + 3 is i1_i1
        else
            intHelper(opcode+1, resultTemplate);  // i_i + 1 is i_i1
    } else if (nx == 1)
        intHelper(opcode+2, resultTemplate);      // i_i + 2 is i1_i
    else if (nx == ny)
        intHelper(opcode, resultTemplate);        // i_i     no change
}

static R_INLINE void realHelper(JIT_OPCODE opcode, int resultLen)
{
    genjit(opcode, R_NilValue, NULL,
           REALSXP, resultLen, R_NilValue, R_NilValue, R_NilValue);
}

// Invoked  for real arith ops like "+".
// This shadows arithmetic.c:real_binary() and is needed because
// real_binary() typecasts both args to real.  We don't want to do
// that, for efficiency in evalJIT() and because we have already
// pushed the args in their original type.

void genjitRealBin(ARITHOP_TYPE oper, CSEXP x, CSEXP y)
{
    static JIT_OPCODE arithToJit[] = { // see ARITHOP_TYPE
        JIT_endop,    // unused
        JIT_add_r_r,  // PLUSOP = 1
        JIT_sub_r_r,  // MINUSOP
        JIT_mul_r_r,  // TIMESOP
        JIT_div_r_r,  // DIVOP
        JIT_pow_r_r,  // POWOP
        JIT_mod_r_r,  // MODOP
        JIT_idiv_r_r  // IDIVOP
    };
    int nx = LENGTH(x);
    int ny = LENGTH(y);
    JIT_OPCODE opcode = arithToJit[oper];

    Dassert(oper > 0 && oper <= IDIVOP);
    Dassert(TYPEOF(x) == REALSXP || TYPEOF(y) == REALSXP);

    if (nx == 0 || ny == 0)
        return;

    if (opcode == JIT_mod_r_r || opcode == JIT_idiv_r_r)
        return;                 // RA_TODO not yet implemented in evalJIT()
    opcode = adjustForTypes(opcode, x, y);

    if (ny == 1) {
        if (nx == 1)
            realHelper(opcode+3, 1);   // r_r + 3 is r1_r1
        else
            realHelper(opcode+1, nx);  // r_r + 1 is r_r1
    } else if (nx == 1)
        realHelper(opcode+2, ny);      // r_r + 2 is r1_r
    else if (nx == ny)
        realHelper(opcode, nx);        // r_r     no change
}

// invoked for math functions with a single argument

void genjitMath1(double(* const func)(double), CSEXP argTemplate,
                 CSEXP ansTemplate)
{
    JIT_OPCODE opcode = JIT_endop;

    switch (TYPEOF(argTemplate)) {
        case LGLSXP:                    // treat logicals as integers
        case INTSXP:
            if (LENGTH(argTemplate) == 1)
                opcode = JIT_math1_i1;
            else
                opcode = JIT_math1_i;
            break;
        case REALSXP:
            if (LENGTH(argTemplate) == 1)
                opcode = JIT_math1_r1;
            else
                opcode = JIT_math1_r;
            break;
        default:
            break;
    }
    if (opcode != JIT_endop)
        genjit(opcode, R_NilValue, func,
               NILSXP, 0, ansTemplate, R_NilValue, R_NilValue);
}

//-----------------------------------------------------------------------------
// Routines for executing jitted code

#if DEBUG_JIT
static int jitEvalFail(char *filename, unsigned nline, char *exp)
{
    printf("istack %d\n", istack);
    if (istack >= 1)
        printfSxp(stack[istack], "stack[istack]  ");
    if (istack >= 2)
        printfSxp(stack[istack-1], "stack[istack-1]");
    printfSxp(R_CurrentExpr, "R_CurrentExpr");
    assertFail(filename, nline, exp);
    return 0;
}
#endif

// specialized assert for the evalJIT, provides some extra details

#if DEBUG_JIT
#define DEvalAsset(exp) (void)((exp) || \
                               jitEvalFail(__FILE__, __LINE__, #exp))
#else
#define DEvalAsset(exp)
#endif

// Pointer to current jit op  i.e. instruction pointer.
// It needs to be global for the assertion code.

static JIT_OP *op;

#define TOS (stack[istack])

// Defs for binops e.g. "x + y" puts x at [istack-1] and y at [istack].
// Remember: "y is top of stack; x is top of stack minus 1".

#define X  (stack[istack-1])
#define Y  (stack[istack])
#define XR (REAL(stack[istack-1]))
#define YR (REAL(stack[istack]))
#define XI (INTEGER(stack[istack-1]))
#define YI (INTEGER(stack[istack]))

static void stackOverflowError(JIT_RECORD * const prec)
{
    error(_("JIT stack overflow evaluating %s"),
          deparseAsShortString(prec->original));
}

static R_INLINE void incStack(JIT_RECORD * const prec)
{
#if DEBUG_JIT
    if (istack >= MAX_STACK_LEN-1)
        stackOverflowError(prec);
#endif
    istack++;
}
static R_INLINE void incStackWithCheck(JIT_RECORD * const prec)
{
    if (istack >= MAX_STACK_LEN-1)
        stackOverflowError(prec);
    istack++;
}

static R_INLINE void DAssertCompatibleTypes(SEXPTYPE xtype, SEXPTYPE ytype)
{
#if DEBUG_JIT
    if (xtype == LGLSXP)
        xtype = INTSXP;
    if (ytype == LGLSXP)
        ytype = INTSXP;
    DEvalAsset(xtype == ytype);
#endif
}

// Check that the pointer to a symbol binding (the operand arg below)
// is the same as that returned by findVarLoc. Also check the value.

#if DEBUG_SYM
static R_INLINE void DCheckSymOpConsistency(CSEXP operand, CSEXP sym,
                                            CSEXP env, CSEXP tos)
{
    SEXP val, findVarVal;
    SEXP findVarLoc1 = findVarLoc(op->sym, op->env);
    if (findVarLoc1 != operand) {
        printfSxp(operand, "\noperand");
        printfSxp(sym, "sym ");
        printfSxp(env, "env ");
        printfSxp(findVarLoc1, "findVarLoc1 ");
        assertFail(__FILE__, __LINE__, "findVarLoc1 != operand");
    }
    val = getSymValFromLoc(operand);
    findVarVal = findVar(sym, env);
    if (val != findVarVal) {
        printfSxp(operand, "\noperand");
        printfSxp(sym, "sym ");
        printfSxp(env, "env ");
        printfSxp(findVarLoc1, "findVarLoc1 ");
        printfSxp(findVarVal, "findVarVal ");
        assertFail(__FILE__, __LINE__, "val != findVarVal");
    }
    assert(tos == R_NilValue || LENGTH(tos) == LENGTH(val));
}
#endif

static R_INLINE void pushSymAux(const JIT_OP *op, CSEXP ans,
                                SEXPTYPE opType, JIT_RECORD * const prec)
{
#if DEBUG_SYM
    DCheckSymOpConsistency(op->operand, op->sym, op->env, R_NilValue);
#endif
    DEvalAsset(TYPEOF(op->operand) == LISTSXP || TYPEOF(op->operand) == SYMSXP);

    TOS = getSymValFromLoc(op->operand);
    if (TYPEOF(TOS) == PROMSXP)
        TOS = PRVALUE(TOS);

    DEvalAsset(TOS != R_UnboundValue && TOS != R_NilValue);
    DAssertCompatibleTypes(TYPEOF(TOS), opType);
}
static R_INLINE void pushSym(const JIT_OP *op, CSEXP ans,
                             SEXPTYPE opType, JIT_RECORD * const prec)
{
    incStack(prec);
    pushSymAux(op, ans, opType, prec);
}
static R_INLINE void pushSymWithCheck(const JIT_OP *op, CSEXP ans,
                            SEXPTYPE opType, JIT_RECORD * const prec)
{
    incStackWithCheck(prec);
    pushSymAux(op, ans, opType, prec);
}
static R_INLINE void DCheckMath1(       // invoked for vec arg to math1
                                 CSEXP result, SEXPTYPE xtype)
{
    DEvalAsset(istack >= 1);
    DAssertCompatibleTypes(TYPEOF(result), REALSXP);
    DAssertCompatibleTypes(TYPEOF(TOS), xtype);
    DEvalAsset(LENGTH(TOS) == LENGTH(result));
}
static R_INLINE void DCheckMath1_1(     // invoked for scalar arg to math1
                                   CSEXP result, SEXPTYPE xtype)
{
    DEvalAsset(istack >= 1);
    DAssertCompatibleTypes(TYPEOF(result), REALSXP);
    DAssertCompatibleTypes(TYPEOF(TOS), xtype);
    DEvalAsset(LENGTH(TOS) == 1);
    DEvalAsset(LENGTH(result) == 1);
}
static R_INLINE void CheckSubscript_i(CSEXP x, int i, JIT_RECORD * const prec)
{
    if (i < 0 || i >= LENGTH(X)) {
        error(_("out of range index %d (allowed range is 1 to %d)\n"
              "       in expression %s"),
            i+1, LENGTH(x), deparseAsShortString(prec->original));
    }
}
static R_INLINE void CheckSubscript_r(CSEXP x, double i,
                                       JIT_RECORD * const prec)
{
    const int ii = (int)i;
    if (ii < 0 || ii >= LENGTH(X)) {
        error(_("out of range index %g (allowed range is 1 to %d)\n"
              "       in expression %s"),
              i+1, LENGTH(x), deparseAsShortString(prec->original));
    }
}
static R_INLINE void DCheckSubscript_x_y1(SEXPTYPE xtype, SEXPTYPE ytype,
                                          CSEXP result)
{
    DEvalAsset(istack >= 2);
    DAssertCompatibleTypes(TYPEOF(X), xtype);
    DAssertCompatibleTypes(TYPEOF(Y), ytype);
    DAssertCompatibleTypes(TYPEOF(result), xtype);
    DEvalAsset(LENGTH(Y) == 1);
    DEvalAsset(LENGTH(result) == 1);
}
static R_INLINE void DCheckAssign(const JIT_OP *op,
                                  SEXPTYPE expectedLhsType,
                                  SEXPTYPE expectedRhsType,
                                  int expectedRhsLen)
{
    DEvalAsset(istack >= 1);
#if DEBUG_SYM
    DCheckSymOpConsistency(op->operand, op->sym, op->env, TOS);
#endif
#if DEBUG_JIT
    {
    SEXP currentVal = getSymValFromLoc(op->operand);
    DAssertCompatibleTypes(TYPEOF(currentVal), expectedLhsType);
    DEvalAsset(LENGTH(currentVal) == LENGTH(Y));
    }
#endif
    DAssertCompatibleTypes(TYPEOF(Y), expectedRhsType);
    DEvalAsset(expectedRhsLen == -1 || LENGTH(Y) == expectedRhsLen);
}
static R_INLINE void DCheckBinop_x_y(SEXPTYPE xtype, SEXPTYPE ytype, CSEXP result)
{
    DEvalAsset(istack >= 2);
    DAssertCompatibleTypes(TYPEOF(X), xtype);
    DAssertCompatibleTypes(TYPEOF(Y), ytype);
    DEvalAsset(LENGTH(X) == LENGTH(Y));
    DEvalAsset(LENGTH(result) == length(X));
}
static R_INLINE void DCheckBinop_x_y1(SEXPTYPE xtype, SEXPTYPE ytype, CSEXP result)
{
    DEvalAsset(istack >= 2);
    DAssertCompatibleTypes(TYPEOF(X), xtype);
    DAssertCompatibleTypes(TYPEOF(Y), ytype);
    DEvalAsset(LENGTH(Y) == 1);
    DEvalAsset(LENGTH(result) == length(X));
}
static R_INLINE void DCheckBinop_x1_y(SEXPTYPE xtype, SEXPTYPE ytype, CSEXP result)
{
    DEvalAsset(istack >= 2);
    DAssertCompatibleTypes(TYPEOF(X), xtype);
    DAssertCompatibleTypes(TYPEOF(Y), ytype);
    DEvalAsset(LENGTH(X) == 1);
    DEvalAsset(LENGTH(result) == length(Y));
}
static R_INLINE void DCheckBinop_x1_y1(SEXPTYPE xtype, SEXPTYPE ytype, CSEXP result)
{
    DEvalAsset(istack >= 2);
    DAssertCompatibleTypes(TYPEOF(X), xtype);
    DAssertCompatibleTypes(TYPEOF(Y), ytype);
    DEvalAsset(LENGTH(X) == 1);
    DEvalAsset(LENGTH(Y) == 1);
    DEvalAsset(LENGTH(result) == 1);
}

// RA_TODO this can return with useless values left on the stack
// e.g. for R code 1;2; which is Cpush_r1; Cpush_r1;
// We should reset istack to 0 at each ";"?

SEXP evalJIT(CSEXP e)       // the JIT code execution machine
{
    JIT_RECORD * const p = (JIT_RECORD *)RAW(e->u.jitsxp.pjit);
    op = p->ops;
    istack = 0;

    DEvalAsset(TYPEOF(e) == JITSXP);
    DEvalAsset(TYPEOF(CAR(e)) == RAWSXP);
    #if DEBUG_JIT >= 2
        printf("# evalJIT %s[%d] expression %s\n", type2char(TYPEOF(p->ans)),
               LENGTH(p->ans), deparseAsShortString(p->original));
    #endif

    while(1) {              // JIT_endop in jitops.c will exit this loop
        double *xr, *yr, *ar, xr0, yr0;     // temps for faster code
        int i, n, *xi, *yi, *ai, xi0, yi0;
        double(*func1)(double);             // math1 function

        SEXP result;

        #if DEBUG_JIT >= 2
            printf("# evalJIT    ");
            printJitOp(op);
        #endif

        switch (op->opcode) {
            #include "jitops.c"
        }
        #if DEBUG_JIT >= 2
            printfSxp(TOS, "# evalJIT\t\t\t\t\t\t\t-->");
        #endif

        op++;
    }
donefor:
    // RA_TODO add NAMED handling, if TYPEOF(p->ans) is a symbol?
    p->ans = stack[istack--];
    #if DEBUG_JIT >= 2
        printfSxp(p->ans, "# evalJIT returns");
        printf("#\n");
    #endif
    return p->ans;
}
